//
// QuickFile Auto Complete Session
// $Id: quickfile-acs.js,v 1.11 2004/11/06 11:22:43 Paul Tomlin Exp $

function QuickFileACS ()
{
	// RDF service
	this.rdfService = Components.classes['@mozilla.org/rdf/rdf-service;1']
						.getService(Components.interfaces.nsIRDFService);
	
	// Message accounts datasource
	this.rdfMsgAccounts = Components.classes['@mozilla.org/rdf/datasource;1?name=msgaccountmanager']
									.createInstance(Components.interfaces.nsIRDFDataSource);
	
	// Mail News RDF datasource
	this.rdfMailNews = Components.classes['@mozilla.org/rdf/datasource;1?name=mailnewsfolders']
							.createInstance(Components.interfaces.nsIRDFDataSource);
	
	// URI to start the search from
	this.searchRootURI = 'msgaccounts:/';
	
	// array of search roots, (possibly) expanded from the searchRootURI
	this.searchRootURIs = new Array();
	
	// compiled RegExp for pattern matching the folder names
	this.re = null;
	
	// are pattern matches case sensitive
	this.caseSensitive = false;
	
	// do we expand spaces in the search string
	this.expandSpaces = true;
	
	// holder for the nsIAutoCompleteResults
	this.results = null;
	
	// holder for the nsIAutoCompleteListener
	this.listener = null;
	
	// callback method to provide the holding code
	//  with the uri for the folder, since there
	//  appears to be no name/value style system
	//  within the autocomplete framework
	this.onAutoCompleteCallback = null;
	
	// debugging?
	this.debug = false;
};

// prepare
// Called to initialise variables which are required to search but
//  don't change between lookups
QuickFileACS.prototype.prepare = function ()
{
	if (this.debug)
		qfDump('QuickFileACS::prepare()');
	
	// get an array of the search root URIs to use
	this.expandSearchRootURI(this.searchRootURI);

	// check that we have something to search within
	// this may be zero if the options don't expand into anything useful
	//  such as msgaccount:/ and an nntp message selected
	if (this.searchRootURIs.length == 0)
	{
		if (this.debug)
			qfDump('QuickFileACS;;prepare - expanding search root failed to produce any usable URIs');
		
		return false;
	}
	
	return true;
};

// onStartLookup
// Called when the user stops typing for a preset duration
QuickFileACS.prototype.onStartLookup = function (searchString, previousSearchResult, listener)
{
	if (this.debug)
		qfDump('QuickFileACS::onStartLookup("' + searchString + '", ...)');
	
	// do the actual searching
	var rv = this.search(searchString);

	// was there an error
	if (! rv)
	{
		if (this.debug)
			qfDump('QuickFileACS::onStartLookup - search returned false, ignoring results');
		
		this.finishAutoCompleteLookup(Components.interfaces.nsIAutoCompleteStatus.ignored);
		
		return;
	}
	
	// keep the listner for later
	this.listener = listener;
	
	// did we find anything
	var status = this.results.items.Count() ?
					Components.interfaces.nsIAutoCompleteStatus.matchFound :
					Components.interfaces.nsIAutoCompleteStatus.noMatch;
	
	this.finishAutoCompleteLookup(status);
};

// onStopLookup
// Called when the user starts typing
QuickFileACS.prototype.onStopLookup = function ()
{
	if (this.debug)
		qfDump('QuickFileACS::onStopLookup()');
	
	// clear up just in case
	this.reset();
};

// onAutoComplete
// Called when the user selects an entry from the popup
QuickFileACS.prototype.onAutoComplete = function (searchString, previousSearchResult, listener)
{
	if (this.debug)
		qfDump('QuickFileACS::onAutoComplete("' + searchString + '", ...)');
	
	// do the actual search
	// this is done yet again because we cannot be sure that the
	// UI has not invalidated our previous results
	var rv = this.search(searchString);
	
	// was there an error
	if (! rv)
	{
		if (this.debug)
			qfDump('QuickFileACS::onAutoComplete - search returned false');
		
		return;
	}
	
	// keep the listner for later
	this.listener = listener;
	
	// holder for the selected URI
	var uri = null;
	
	// find the selected uri from the last search result
	for (var i = 0; i < this.results.items.Count(); i++)
	{
		var result = this.results.items.GetElementAt(i);
		
		if (result.value == searchString)
		{
			uri = result.param.URI;
			break;
		}
	}
	
	if (! uri)
	{
		// incomplete lookup, user didn't select anything useful
		if (this.debug)
			qfDump('QuickFileACS::onAutoComplete - no URI found for a folder named "' + searchString + '"');
		
		return;
	}
		
	// if we have a callback to use then call it
	if (typeof this.onAutoCompleteCallbackFunc == 'function')
	{
		this.onAutoCompleteCallbackFunc.apply(this.onAutoCompleteCallbackObj, [result.value, uri]);
	}
};

// expandSearchRootURI
// Called to possibly expand the supplied search root into an array of many
QuickFileACS.prototype.expandSearchRootURI = function (searchRootURI)
{
	if (this.debug)
		qfDump('QuickFileACS::expandSearchRootURI("' + searchRootURI + '")');
	
	switch (searchRootURI)
	{
		// if we are searching all accounts, we need to query the account manager
		case 'msgaccounts:/':
			var accountsRoot = this.rdfService.GetResource(searchRootURI);
			var childArc = this.rdfService.GetResource('http://home.netscape.com/NC-rdf#child');
			var accounts = this.rdfMsgAccounts.GetTargets(accountsRoot, childArc, true);
			while (accounts.hasMoreElements())
			{
				var account = accounts.getNext();
				try
				{
					var searchRoot = account.QueryInterface(Components.interfaces.nsIMsgFolder);
					
					this.addSearchRootURI(searchRoot.URI);
				}
				catch (e)
				{
					// for some reason the account did not yield an nsIMsgFolder
					if (this.debug)
						qfDump('QuickFileACS::onStartLookup - failed to get an nsIMsgFolder, ignoring this node');
				}
			}
		break;
		
		// if we want the root folder of the selected message, find that
		case 'msgaccount:/':
			// get the URI of the first selected message
			var message = window.opener.GetSelectedMessages()[0];
			
			// handle NNTP messages specifically in this case
			if (message.substr(0, 14) == 'news-message:/')
			{
				// don't expand this as we cannot move news messages within
				//  the server.. 
				break;
			}
			
			// get the nsIMsgFolder from the message URI
			var folder = qfGetFolderFromMessageURI(message);
			
			// if it looks real
			if (folder)
			{
				// get the root of that tree
				folder = folder.rootFolder;
				
				this.addSearchRootURI(folder.URI);
			}
			else
			{
				if (this.debug)
					qfDump('QuickFileACS::expandSearchRootURI - failed to get message folder for ' + message);
			}
		break;
		
		// otherwise just use the provided URI
		default:
			this.addSearchRootURI(searchRootURI);
		break;
	}
};

// addSearchRootURI
// Called to add a node from which to start searching for matching folders
QuickFileACS.prototype.addSearchRootURI = function (searchRootURI)
{
//	qfDump('QuickFileACS::addSearchRoot("' + searchRootURI + '")');
	
	// we can't move messages into news:// folders, so don't search them
	if (searchRootURI.substr(0, 7) == 'news://')
	{
		if (this.debug)
			qfDump('QuickFileACS::addSearchRoot - ignoring news folder ' + searchRootURI);
		
		return false;
	}
	
//	qfDump('QuickFileACS::addSearchRoot - adding ' + searchRootURI);
	this.searchRootURIs.push(searchRootURI);
};

// search
// Search from the 'search root'
QuickFileACS.prototype.search = function (searchString)
{
	if (this.debug)
		qfDump('QuickFileACS::search("' + searchString + '")');

	// ignore empty search strings
	// TODO: cancelled for now, maybe an option?
	if (searchString == '' && false)
	{
		if (this.debug)
			qfDump('QuickFileACS::search - ignoring empty search string');
		
		return false;
	}

	// create a new results instance to return to the listener
	this.results = Components.classes['@mozilla.org/autocomplete/results;1']
					.createInstance(Components.interfaces.nsIAutoCompleteResults);

	// keep the search string for later incantations of search
	this.results.searchString = searchString;

	// compile a new RegExp for comparisons
	var reOpts = this.caseSensitive ? '' : 'i';
	var reString = this.expandSpaces ? searchString.split(/\s+/).join('.+') : searchString;

	this.re = new RegExp(reString, reOpts);

	for (var i = 0; i < this.searchRootURIs.length; i++)
	{
		var searchRootURI = this.searchRootURIs[i];
		
//		qfDump('QuickFileACS::search - starting search of ' + searchRootURI);
		
		var folder = null;
		try
		{
			// get an nsIMsgFolder from the URI
			var folder = this.rdfService.GetResource(searchRootURI)
				.QueryInterface(Components.interfaces.nsIMsgFolder);
		}
		catch (e)
		{
			// for some reason failed to get an nsIMsgFolder from the URI
			if (this.debug)
				qfDump('QuickFileACS::search - failed to get an nsIMsgFolder from ' + searchRootURI + "\n" + e);
				
			next;
		}
		
		if (folder)
			this.searchFolder(folder);
	}
	
	return true;
};

// searchFolder
// Recursively check if the folder or any children match the search string
// prefix stores the parent folders full name
QuickFileACS.prototype.searchFolder = function (folder, prefix)
{
//	if (this.debug)
//		qfDump('QuickFileACS::searchFolder(...)');
	
	// ensure we have got a valid nsIMsgFolder
	if (! (folder instanceof Components.interfaces.nsIMsgFolder))
	{
		if (this.debug)
			qfDump('QuickFileACS::searchFolder - parameter is not an nsIMsgFolder');
			
		return;
	}
	
	// get the fully qualified name of the folder
	var fullname = (prefix ? prefix + '/' : '') + folder.name;
	
	if (this.isValidMatch(folder, fullname))
	{
		this.addResult(folder, fullname);
	}
	
	try
	{
		var node = folder.QueryInterface(Components.interfaces.nsIRDFNode);
						
		var childArc = this.rdfService.GetResource('http://home.netscape.com/NC-rdf#child');
		
		var children = this.rdfMailNews.GetTargets(node, childArc, true);
		while (children.hasMoreElements())
		{
			var child = children.getNext()
						.QueryInterface(Components.interfaces.nsIMsgFolder);
			
			this.searchFolder(child, fullname);
		}
	}
	catch (e)
	{
		// error trying to get the child folders
		if (this.debug)
			qfDump('QuickFileACS::searchFolder - failed to query children from ' + _folder.URI + "\n" + e);
			
		return false;
	}
};

// isValidMatch
// Called to check if the folder matches the search string and is a valid
//  result (canFileMessages)
// fullname is a string of the parent folder name
QuickFileACS.prototype.isValidMatch = function (folder, fullname)
{
//	if (this.debug)
//		qfDump('QuickFileACS::isValidMatch(...)');

	if (! (folder instanceof Components.interfaces.nsIMsgFolder))
	{
		if (this.debug)
			qfDump('QuickFileACS::isValidMatch - not an nsIMsgFolder ' + folder);
			
		return false;
	}	

	if (! this.re.exec(fullname))
	{
//		if (this.debug)
//			qfDump('QuickFileACS::isValidMatch - search string not found in ' + fullname);
		return false;
	}
	
	if (! folder.canFileMessages)
	{
//		if (this.debug)
//			qfDump('QuickFileACS::isValidMatch - cannot file messages in ' + fullname);
		return false;
	}

//	if (this.debug)
//		qfDump('QuickFileACS::isValidMatch - found valid folder ' + fullname);
	
	return true;
};

// addResult
// Called to add the folder to the result set. Also adds an entry into the
//  label -> uri lookup table
QuickFileACS.prototype.addResult = function (folder, fullname)
{
//	if (this.debug)
//		qfDump('QuickFileACS::addResult("<' + folder.URI + '>")');
	
	var result = Components.classes['@mozilla.org/autocomplete/item;1']
					.createInstance(Components.interfaces.nsIAutoCompleteItem);
	
	result.value = fullname;
	result.comment = folder.URI;
	result.param = folder;
	
	this.results.items.AppendElement(result);
};

// finishAutoCompleteLookup
// Called to send the results/error back to the textbox
QuickFileACS.prototype.finishAutoCompleteLookup = function (status)
{
//	if (this.debug)
//		qfDump('QuickFileACS::finishAutoCompleteLookup(' + status + ')');
	
	switch (status)
	{
		// a match was found, send it
		case Components.interfaces.nsIAutoCompleteStatus.matchFound:
			this.listener.onAutoComplete(this.results, status);
			break;

		// there was an error but a message needs to be sent in the
		//  results
		case Components.interfaces.nsIAutoCompleteStatus.failureItems:
			// TODO: handle this
			//break;

		// no results, no interest
		case Components.interfaces.nsIAutoCompleteStatus.ignored:
		case Components.interfaces.nsIAutoCompleteStatus.noMatch:
		case Components.interfaces.nsIAutoCompleteStatus.failed:
			this.listener.onAutoComplete(null, status);
			break;
	}
	
	// clean up
	this.reset();
};

// reset
// Called to clear all intra-search variables
QuickFileACS.prototype.reset = function ()
{
	if (this.debug)
		qfDump('QuickFileACS::reset');
		
	this.listner = null;
	this.results = null;
	this.re = null;
};
